﻿Imports System
Imports System.Collections.Generic
Imports System.Text
Imports System.IO
Imports System.Collections

Namespace TabuMaxClique
    Module TabuMaxCliqueProgram
        Dim random As Random = Nothing

        Public Sub Main(args As String())
            Try
                Console.WriteLine(vbLf & "Begin tabu algorithm maximum clique demo" & vbLf)

                Dim graphFile As String = "DimacsGraph.clq"
                Console.WriteLine("Graph data DIMACS file is " + graphFile)
                Console.WriteLine("Loading graph into memory")
                Console.WriteLine("Graph loaded and validated" & vbLf)

                Dim graph As New MyGraph(graphFile, "DIMACS")

                Dim maxTime As Integer = 50 ' for demo program
                'Dim maxTime As Integer = 100000  ' for real problems
                Dim targetCliqueSize As Integer = graph.NumberNodes

                Console.WriteLine("Using tabu algorithm with adaptive prohibit time" & vbLf)
                Dim maxClique As List(Of Integer) = FindMaxClique(graph, maxTime, targetCliqueSize)
                Console.WriteLine(vbLf & vbLf & "Maximum time reached")
                Console.WriteLine(vbLf & "Size of best clique found = " & maxClique.Count.ToString())

                Console.WriteLine(vbLf & "Best clique found:")
                Console.WriteLine(ListAsString(maxClique))

                Console.WriteLine(vbLf & "End tabu algorithm maximum clique demo" & vbLf)
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine("Fatal: " & ex.Message)
                Console.ReadLine()
            End Try
        End Sub ' Main

        Private Function FindMaxClique(graph As MyGraph, maxTime As Integer, targetCliqueSize As Integer) As List(Of Integer)
            Dim clique As New List(Of Integer)()
            random = New Random(1)
            Dim time As Integer = 0
            Dim timeBestClique As Integer = 0
            Dim timeRestart As Integer = 0
            Dim prohibitPeriod As Integer = 1
            Dim timeProhibitChanged As Integer = 0
            Dim nodeToAdd As Integer = -1
            Dim nodeToDrop As Integer = -1

            Dim lastMoved As Integer() = New Integer(graph.NumberNodes - 1) {}
            ' the time when [node] was last added or dropped; determines tabu or not.
            For i As Integer = 0 To lastMoved.Length - 1
                lastMoved(i) = Integer.MinValue
            Next

            Dim history As New Hashtable()
            ' cliques seen. used to modify prohibit period
            Dim hs As System.Collections.Generic.HashSet(Of CliqueInfo) = New HashSet(Of CliqueInfo)()
            Dim randomNode As Integer = random.[Next](0, graph.NumberNodes)
            Console.WriteLine("Adding node " & randomNode.ToString())
            clique.Add(randomNode)

            Dim bestClique As New List(Of Integer)()
            bestClique.AddRange(clique)
            Dim bestSize As Integer = bestClique.Count
            timeBestClique = time

            Dim possibleAdd As List(Of Integer) = MakePossibleAdd(graph, clique)
            ' nodes which will increase size of clique
            Dim oneMissing As List(Of Integer) = MakeOneMissing(graph, clique)
            ' used to determine best node to drop
            While time < maxTime AndAlso bestSize < targetCliqueSize
                time += 1

                Dim cliqueChanged As Boolean = False
                ' to control branching logic in loop
                If possibleAdd.Count > 0 Then
                    ' 
                    Dim allowedAdd As List(Of Integer) = SelectAllowedNodes(possibleAdd, time, prohibitPeriod, lastMoved)
                    ' could have Count == 0
                    If allowedAdd.Count > 0 Then
                        nodeToAdd = GetNodeToAdd(graph, allowedAdd, possibleAdd)
                        ' node from allowedAdd which is most connected to nodes in possibleAdd
                        Console.WriteLine("Adding node " & nodeToAdd.ToString())
                        clique.Add(nodeToAdd)
                        lastMoved(nodeToAdd) = time
                        clique.Sort()
                        cliqueChanged = True
                        If clique.Count > bestSize Then
                            bestSize = clique.Count
                            bestClique.Clear()
                            bestClique.AddRange(clique)
                            Console.WriteLine("new best = " & bestSize.ToString())
                            timeBestClique = time
                        End If
                    End If
                End If
                ' add allowed node
                If cliqueChanged = False Then
                    If clique.Count > 0 Then
                        Dim allowedInClique As List(Of Integer) = SelectAllowedNodes(clique, time, prohibitPeriod, lastMoved)
                        ' nodes which are in clique and which are allowed
                        If allowedInClique.Count > 0 Then
                            nodeToDrop = GetNodeToDrop(graph, allowedInClique, oneMissing)
                            ' find allowed node in clique which generate max increase in possibleAdd set.
                            Console.WriteLine("Dropping allowed node " & nodeToDrop.ToString())
                            clique.Remove(nodeToDrop)
                            lastMoved(nodeToDrop) = time
                            clique.Sort()
                            cliqueChanged = True
                        End If
                    End If
                End If
                ' drop allowed node
                If cliqueChanged = False Then
                    If clique.Count > 0 Then
                        nodeToDrop = clique(random.[Next](0, clique.Count))
                        Console.WriteLine("Dropping random node " & nodeToDrop.ToString())
                        clique.Remove(nodeToDrop)
                        lastMoved(nodeToDrop) = time
                        clique.Sort()
                        cliqueChanged = True
                    End If
                End If
                ' drop random node
                'Dim restart As Integer = 100 * bestSize ' for real problems
                Dim restart As Integer = 2 * bestSize
                ' dummy short value for demo program
                If time - timeBestClique > restart AndAlso time - timeRestart > restart Then
                    ' restart?
                    Console.WriteLine(vbLf & "Restarting with prohibit period " & prohibitPeriod.ToString())
                    timeRestart = time
                    prohibitPeriod = 1
                    timeProhibitChanged = time

                    history.Clear()

                    Dim seedNode As Integer = -1
                    Dim temp As New List(Of Integer)()
                    For i As Integer = 0 To lastMoved.Length - 1
                        If lastMoved(i) = Integer.MinValue Then
                            temp.Add(i)
                        End If
                    Next
                    If temp.Count > 0 Then
                        seedNode = temp(random.[Next](0, temp.Count))
                    Else
                        seedNode = random.[Next](0, graph.NumberNodes)
                    End If

                    clique.Clear()
                    Console.WriteLine("Adding seed node " & seedNode.ToString())
                    clique.Add(seedNode)
                End If
                ' restart
                possibleAdd = MakePossibleAdd(graph, clique)
                oneMissing = MakeOneMissing(graph, clique)

                prohibitPeriod = UpdateProhibitPeriod(graph, clique, bestSize, history, time, prohibitPeriod, _
                 timeProhibitChanged)
            End While
            ' main processing loop
            Return bestClique
        End Function ' FindMaxClique

        Private Function MakePossibleAdd(graph As MyGraph, clique As List(Of Integer)) As List(Of Integer)
            ' create list of nodes in graph which are connected to all nodes in clique and therefore will form a larger clique
            ' calls helper FormsALargerClique
            Dim result As New List(Of Integer)()

            For i As Integer = 0 To graph.NumberNodes - 1
                ' each node in graph
                If FormsALargerClique(graph, clique, i) = True Then
                    result.Add(i)
                End If
            Next
            Return result
            ' could be empty List
        End Function ' MakePossibleAdd

        Private Function FormsALargerClique(graph As MyGraph, clique As List(Of Integer), node As Integer) As Boolean
            ' is node connected to all nodes in clique?
            For i As Integer = 0 To clique.Count - 1
                ' compare node against each member of clique
                'If (clique(i) = node) Then Return False 'node is already in clique so node will not form a larger clique
                If graph.AreAdjacent(clique(i), node) = False Then
                    Return False
                    ' node is not connected to one of the nodes in the clique
                End If
            Next
            Return True
            ' passed all checks
        End Function ' FormsALargerClique

        Private Function GetNodeToAdd(graph As MyGraph, allowedAdd As List(Of Integer), possibleAdd As List(Of Integer)) As Integer
            ' find node from a List of allowedAdd which has max degree in posibleAdd
            ' there could be more than one, if so, pick one at random
            'If (IsNothing(allowedAdd)) Then Throw New Exception("List allowedPossibleAdd is null in GetNodeToAdd")
            'If (IsNothing(possibleAdd)) Then Throw New Exception("List possibleAdd is null in GetNodeToAdd")
            'If (allowedAdd.Count = 0) Then Throw New Exception("List allowedPossibleAdd has Count 0 in GetNodeToAdd")
            'If (possibleAdd.Count = 0) Then Throw New Exception("List possibleAdd has Count 0 in GetNodeToAdd")

            If allowedAdd.Count = 1 Then
                ' there is only 1 node to choose from
                Return allowedAdd(0)
            End If

            ' examine each node in allowedAdd to find the max degree in possibleAdd (because there could be several nodes tied with max degree)
            Dim maxDegree As Integer = 0
            For i As Integer = 0 To allowedAdd.Count - 1
                Dim currNode As Integer = allowedAdd(i)
                Dim degreeOfCurrentNode As Integer = 0
                For j As Integer = 0 To possibleAdd.Count - 1
                    ' check each node in possibleAdd
                    Dim otherNode As Integer = possibleAdd(j)
                    If graph.AreAdjacent(currNode, otherNode) = True Then
                        degreeOfCurrentNode += 1
                    End If
                Next
                If degreeOfCurrentNode > maxDegree Then
                    maxDegree = degreeOfCurrentNode
                End If
            Next

            ' now rescan allowedAdd and grab all nodes which have maxDegree
            Dim candidates As New List(Of Integer)()
            For i As Integer = 0 To allowedAdd.Count - 1
                Dim currNode As Integer = allowedAdd(i)
                Dim degreeOfCurrentNode As Integer = 0
                For j As Integer = 0 To possibleAdd.Count - 1
                    ' check each node in possibleAdd
                    Dim otherNode As Integer = possibleAdd(j)
                    If graph.AreAdjacent(currNode, otherNode) = True Then
                        degreeOfCurrentNode += 1
                    End If
                Next
                If degreeOfCurrentNode = maxDegree Then
                    candidates.Add(currNode)
                End If
            Next

            'If (candidates.Count = 0) Then Throw New Exception("candidates List has size 0 just before return in GetNodeToAdd")
            Return candidates(random.[Next](0, candidates.Count))
            ' if candidates has Count 1 we'll get that one node
        End Function ' GetNodeToAdd

        Private Function GetNodeToDrop(graph As MyGraph, allowedInClique As List(Of Integer), oneMissing As List(Of Integer)) As Integer
            ' get a node from allowed clique set, which if dropped, gives the largest increase in PA set size
            ' we use the oneMissing set to determine which clique node to pick
            'If (IsNothing(clique)) Then Throw New Exception("clique is null in GetNodeToDrop")
            'If (clique.Count = 0) Then Throw New Exception("clique has Count 0 in GetNodeToDrop")

            If allowedInClique.Count = 1 Then
                Return allowedInClique(0)
            End If

            ' scan each node in clique and determine the max possibleAdd size if node removed
            Dim maxCount As Integer = 0
            ' see explanation below
            For i As Integer = 0 To allowedInClique.Count - 1
                ' each node in clique nodes List
                Dim currCliqueNode As Integer = allowedInClique(i)
                Dim countNotAdjacent As Integer = 0
                For j As Integer = 0 To oneMissing.Count - 1
                    ' each node in the one missing list
                    Dim currOneMissingNode As Integer = oneMissing(j)
                    If graph.AreAdjacent(currCliqueNode, currOneMissingNode) = False Then
                        ' we like this
                        countNotAdjacent += 1

                        ' if currCliqueNode is not connected to omNode then currCliqueNode is the 'missing'
                        ' it would be good to drop this cliqueNode because after dropped from clique
                        ' the remaining nodes in the clique will all be connected to the omNode
                        ' and so the omNode would become a posibleAdd node and increase PA set size
                        ' So the best node to drop from clique will be the one which is least connected
                        ' to the nodes in OM
                    End If
                Next

                If countNotAdjacent > maxCount Then
                    maxCount = countNotAdjacent
                End If
            Next

            ' at this point we know what the max-not-connected count is but there could be several clique nodes which give that size
            Dim candidates As New List(Of Integer)()
            For i As Integer = 0 To allowedInClique.Count - 1
                ' each node in clique
                Dim currCliqueNode As Integer = allowedInClique(i)
                Dim countNotAdjacent As Integer = 0
                For j As Integer = 0 To oneMissing.Count - 1
                    ' each node in the one missing list
                    Dim currOneMissingNode As Integer = oneMissing(j)
                    If graph.AreAdjacent(currCliqueNode, currOneMissingNode) = False Then
                        countNotAdjacent += 1
                    End If
                Next

                If countNotAdjacent = maxCount Then
                    ' current clique node has max count not connected
                    candidates.Add(currCliqueNode)
                End If
            Next

            'If (candidates.Count = 0) Then Throw New Exception("candidates List has size 0 just before return in GetNodeToDropFromAllowedInClique")
            Return candidates(random.[Next](0, candidates.Count))
            ' must have size of at least 1
        End Function ' GetNodeToDrop

        Private Function MakeOneMissing(graph As MyGraph, clique As List(Of Integer)) As List(Of Integer)
            ' make a list of nodes in graph which are connected to all but one of the nodes in clique
            Dim count As Integer
            ' number of nodes in clique which are connected to a candidate node. if final count == (clique size - 1) then candidate is a winner
            Dim result As New List(Of Integer)()

            For i As Integer = 0 To graph.NumberNodes - 1
                ' each node in graph i a candidate
                count = 0
                If graph.NumberNeighbors(i) < clique.Count - 1 Then
                    Continue For
                End If
                ' node i has too few neighbors to possibly be connected to all but 1 node in clique
                'If (LinearSearch(clique, i) = True) Then Continue For
                If clique.BinarySearch(i) >= 0 Then
                    Continue For
                End If
                For j As Integer = 0 To clique.Count - 1
                    ' count how many nodes in clique are connected to candidate i
                    If graph.AreAdjacent(i, clique(j)) Then
                        count += 1
                    End If
                Next
                If count = clique.Count - 1 Then
                    result.Add(i)
                End If
            Next
            ' each candidate node i
            Return result

        End Function ' MakeOneMissing

        Private Function SelectAllowedNodes(listOfNodes As List(Of Integer), time As Integer, prohibitPeriod As Integer, lastMoved As Integer()) As List(Of Integer)
            'If (IsNothing(listOfNodes)) Then Throw New Exception("null listOfNodes in SelectAllowedNodes")
            Dim result As New List(Of Integer)()
            If listOfNodes.Count = 0 Then
                Return result
            End If
            ' if the basis List has no nodes the subset List can't have any nodes
            For i As Integer = 0 To listOfNodes.Count - 1
                Dim currNode As Integer = listOfNodes(i)
                If time > lastMoved(currNode) + prohibitPeriod Then
                    ' allowed because time is greater than time last moved plus the prohibit
                    result.Add(currNode)
                End If
            Next
            Return result
            ' could have Count == 0
        End Function ' SelectAllowedNodes

        Private Function UpdateProhibitPeriod(graph As MyGraph, clique As List(Of Integer), bestSize As Integer, history As Hashtable, time As Integer, prohibitPeriod As Integer, _
         ByRef timeProhibitChanged As Integer) As Integer
            ' increase or decrease the prohibition interval based on whether or not the current clique has been seen
            ' before, and if seen, how recent
            Dim result As Integer = prohibitPeriod
            ' default return is no change
            Dim cliqueInfo As New CliqueInfo(clique, time)
            If history.Contains(cliqueInfo.GetHashCode()) = True Then
                ' current clique is in the history hash table and so has been seen before
                Dim ci As CliqueInfo = DirectCast(history(cliqueInfo.GetHashCode()), CliqueInfo)
                ' to simplify
                Dim intervalSinceLastVisit As Integer = time - ci.LastSeen
                ' how long has it been since the clique was last seen?
                ci.LastSeen = time
                If intervalSinceLastVisit < 2 * (graph.NumberNodes - 1) Then
                    ' it has been a 'short' time since we last saw current clique (we're getting duplicates) 
                    'so increase T to make more prohibited nodes which will decrease duplications of clique.
                    ' increasing the prohibit time
                    timeProhibitChanged = time
                    If prohibitPeriod + 1 < 2 * bestSize Then
                        ' 2 * bestSize or something else
                        Return prohibitPeriod + 1
                    Else
                        ' new prohibit period
                        Return 2 * bestSize
                        ' maximum prohibit period
                    End If
                End If
            Else
                ' currentClique is not in hashTable and therefore has not been seen before
                ' add currentClique and time to hashTable
                history.Add(cliqueInfo.GetHashCode(), cliqueInfo)
            End If

            ' at this point, either the clique has been seen before but was seen a 'long' time ago (so we didn't increase prohibit time)
            ' or clique hasn't been seen before.
            ' if it has been a long time since we made a change in prohibit period then we want to try and reduce it
            ' because we will have a greater number of allowed nodes to choose from.

            If (time - timeProhibitChanged) > (10 * bestSize) Then
                ' we have not changed prohibitPeriod in a long time; let's decrease it to get more allowed nodes
                timeProhibitChanged = time
                If prohibitPeriod - 1 > 1 Then
                    Return prohibitPeriod - 1
                Else
                    Return 1
                    ' minimum prohibit period
                End If
            Else
                ' the default value which is no change
                Return result
            End If

        End Function ' UpdateprohibitPeriod

        Private Function ListAsString(list As List(Of Integer)) As String
            Dim s As String = ""
            For i As Integer = 0 To list.Count - 1
                If i Mod 10 = 0 AndAlso i > 0 Then
                    s += Environment.NewLine
                End If
                s += list(i) & " "
            Next

            s += Environment.NewLine
            Return s
        End Function ' ListAsString

        Private Sub ValidateState(graph As MyGraph, clique As List(Of Integer), possibleAdd As List(Of Integer), oneMissing As List(Of Integer))
            ' any duplicates in clique?
            For i As Integer = 0 To clique.Count - 2
                For j As Integer = i + 1 To clique.Count - 1
                    If clique(i) = clique(j) Then
                        Throw New Exception("Dup value in clique")
                    End If
                Next
            Next
            ' any values in clique in possibleAdd or oneMissing?
            For i As Integer = 0 To possibleAdd.Count - 1
                If clique.BinarySearch(possibleAdd(i)) >= 0 Then
                    Throw New Exception("Possible Add value in clique")
                End If
            Next
            For i As Integer = 0 To oneMissing.Count - 1
                If clique.BinarySearch(oneMissing(i)) >= 0 Then
                    Throw New Exception("One Missing value in clique")
                End If
            Next
            ' any values in possibleAdd in oneMissing?
            For i As Integer = 0 To possibleAdd.Count - 1
                If oneMissing.Contains(possibleAdd(i)) Then
                    Throw New Exception("Value in possibleAdd found in oneMissing")
                End If
            Next
            For i As Integer = 0 To oneMissing.Count - 1
                If possibleAdd.Contains(oneMissing(i)) Then
                    Throw New Exception("Value in oneMissing found in possibleAdd")
                End If
            Next
        End Sub ' ValidateState

        Private Class CliqueInfo
            Private clique As List(Of Integer)
            Private m_lastSeen As Integer

            Public Sub New(clique As List(Of Integer), lastSeen As Integer)
                Me.clique = New List(Of Integer)()
                Me.clique.AddRange(clique)
                Me.m_lastSeen = lastSeen
            End Sub

            Public Property LastSeen() As Integer
                Get
                    Return Me.m_lastSeen
                End Get
                Set(value As Integer)
                    Me.m_lastSeen = value
                End Set
            End Property

            Public Overrides Function GetHashCode() As Integer
                Dim sb As New StringBuilder()
                For i As Integer = 0 To clique.Count - 1
                    sb.Append(clique(i))
                    sb.Append(" ")
                Next

                Dim s As String = sb.ToString()
                Return s.GetHashCode()
            End Function

            Public Overrides Function ToString() As String
                Dim s As String = ""
                For i As Integer = 0 To clique.Count - 1
                    s += clique(i) + " "
                Next
                Return s
            End Function

        End Class         ' class CliqueInfo

    End Module     ' module TabuMaxCliqueProgram

    Public Class MyGraph
        Private data As BitMatrix
        Private m_numberNodes As Integer
        Private m_numberEdges As Integer
        Private m_numberNeighbors As Integer()

        Public Sub New(graphFile As String, fileFormat As String)
            If fileFormat.ToUpper() = "DIMACS" Then
                LoadDimacsFormatGraph(graphFile)
            Else
                Throw New Exception("Format " + fileFormat + " not supported")
            End If
        End Sub

        Private Sub LoadDimacsFormatGraph(graphFile As String)
            Dim ifs As New FileStream(graphFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim tokens As String() = Nothing

            ' advance to and get the p line (ex: "p edge 9 16")
            line = sr.ReadLine()
            ' read first line of file as a priming read
            line = line.Trim()
            While line IsNot Nothing AndAlso line.StartsWith("p") = False
                line = sr.ReadLine()
                line = line.Trim()
            End While

            tokens = line.Split(" "c)
            Dim numNodes As Integer = Integer.Parse(tokens(2))
            ' number nodes
            Dim numEdges As Integer = Integer.Parse(tokens(3))
            ' number edges
            sr.Close()
            ifs.Close()

            Me.data = New BitMatrix(numNodes)

            ifs = New FileStream(graphFile, FileMode.Open)
            ' reopen file
            sr = New StreamReader(ifs)
            While (InlineAssignHelper(line, sr.ReadLine())) IsNot Nothing
                line = line.Trim()
                If line.StartsWith("e") = True Then
                    ' (ex: "e 7 4")
                    tokens = line.Split(" "c)
                    Dim nodeA As Integer = Integer.Parse(tokens(1)) - 1
                    ' DIMACS is 1-based. subtract 1 to convert to 0-based
                    Dim nodeB As Integer = Integer.Parse(tokens(2)) - 1
                    data.SetValue(nodeA, nodeB, True)
                    data.SetValue(nodeB, nodeA, True)
                End If
            End While
            sr.Close()
            ifs.Close()

            Me.m_numberNeighbors = New Integer(numNodes - 1) {}
            For row As Integer = 0 To numNodes - 1
                Dim count As Integer = 0
                For col As Integer = 0 To numNodes - 1
                    If data.GetValue(row, col) = True Then
                        count += 1
                    End If
                Next
                m_numberNeighbors(row) = count
            Next

            Me.m_numberNodes = numNodes
            Me.m_numberEdges = numEdges
            Return
        End Sub

        Public ReadOnly Property NumberNodes() As Integer
            Get
                Return Me.m_numberNodes
            End Get
        End Property

        Public ReadOnly Property NumberEdges() As Integer
            Get
                Return Me.m_numberEdges
            End Get
        End Property

        Public Function NumberNeighbors(node As Integer) As Integer
            Return Me.m_numberNeighbors(node)
        End Function

        Public Function AreAdjacent(nodeA As Integer, nodeB As Integer) As Boolean
            If Me.data.GetValue(nodeA, nodeB) = True Then
                Return True
            Else
                Return False
            End If
        End Function

        Public Overrides Function ToString() As String
            Dim s As String = ""
            For i As Integer = 0 To Me.data.[Dim] - 1
                s += i + ": "
                For j As Integer = 0 To Me.data.[Dim] - 1
                    If Me.data.GetValue(i, j) = True Then
                        s += j + " "
                    End If
                Next
                s += Environment.NewLine
            Next
            Return s
        End Function

        Public Sub ValidateGraphFile(graphFile As String, fileFormat As String)
            If fileFormat.ToUpper() = "DIMACS" Then
                ValidateDimacsGraphFile(graphFile)
            Else
                Throw New Exception("Format " + fileFormat + " not supported")
            End If
        End Sub

        Public Sub ValidateDimacsGraphFile(graphFile As String)
            Dim ifs As New FileStream(graphFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim tokens As String() = Nothing

            While (InlineAssignHelper(line, sr.ReadLine())) IsNot Nothing
                line = line.Trim()
                If line.StartsWith("c") = False AndAlso line.StartsWith("p") = False AndAlso line.StartsWith("e") = False Then
                    Throw New Exception("Unknown line type: " + line + " in ValidateDimacsGraphFile")
                End If

                Try
                    If line.StartsWith("p") Then
                        tokens = line.Split(" "c)
                        Dim numNodes As Integer = Integer.Parse(tokens(2))
                        Dim numEdges As Integer = Integer.Parse(tokens(3))
                    ElseIf line.StartsWith("e") Then
                        tokens = line.Split(" "c)
                        Dim nodeA As Integer = Integer.Parse(tokens(1))
                        Dim nodeB As Integer = Integer.Parse(tokens(2))
                    End If
                Catch
                    Throw New Exception("Error parsing line = " + line + " in ValidateDimacsGraphFile")
                End Try
            End While

            sr.Close()
            ifs.Close()
            Return
        End Sub ' ValidateDimacsGraphFile

        Public Sub ValidateGraph()
            ' total number edges items must be an even number
            Dim numConnections As Integer = 0
            For i As Integer = 0 To Me.data.[Dim] - 1
                For j As Integer = 0 To Me.data.[Dim] - 1
                    If Me.data.GetValue(i, j) = True Then
                        numConnections += 1
                    End If
                Next
            Next
            If numConnections Mod 2 <> 0 Then
                Throw New Exception("Total number of connections in graph is " + numConnections + ". Should be even")
            End If

            ' fully symmetric
            For i As Integer = 0 To Me.data.[Dim] - 1
                If Me.data.GetValue(i, i) = True Then
                    Throw New Exception("Node " + i + " is connected to itself")
                End If
                For j As Integer = 0 To Me.data.[Dim] - 1
                    If Me.data.GetValue(i, j) <> Me.data.GetValue(j, i) Then
                        Throw New Exception("Graph is not symmetric at " + i + " and " + j)
                    End If
                Next
            Next
            Return
        End Sub ' ValidateGraph
        ' ---------------------------------------------------------------------------------------------------------------------------------------------------------

        Private Class BitMatrix
            Private data As BitArray()
            Public ReadOnly [Dim] As Integer

            Public Sub New(n As Integer)
                Me.data = New BitArray(n - 1) {}
                For i As Integer = 0 To data.Length - 1
                    Me.data(i) = New BitArray(n)
                Next
                Me.[Dim] = n
            End Sub
            Public Function GetValue(row As Integer, col As Integer) As Boolean
                Return data(row)(col)
            End Function
            Public Sub SetValue(row As Integer, col As Integer, value As Boolean)
                data(row)(col) = value
            End Sub
            Public Overrides Function ToString() As String
                Dim s As String = ""
                For i As Integer = 0 To data.Length - 1
                    For j As Integer = 0 To data(i).Length - 1
                        If data(i)(j) = True Then
                            s += "1 "
                        Else
                            s += "0 "
                        End If
                    Next
                    s += Environment.NewLine
                Next
                Return s
            End Function

        End Class ' class BitMatrix

        Private Function InlineAssignHelper(Of T)(ByRef target As T, value As T) As T
            target = value
            Return value
        End Function

    End Class ' class MyGraph
End Namespace
